/*
 * borgmove.c
 */
#include "copyright.h"

#include <X11/Xlib.h>
#include <stdio.h>
#include <signal.h>
#include <values.h>
#include <math.h>
#include <time.h>
#include "defs.h"
#include "struct.h"
#include "data.h"

#ifndef _NO_DEPEND
#include "borg_protos.h"
#endif

#include "planets.h"

#ifdef USE_SYS_PROTOS
#if defined(__STDC__) || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

extern int                    printf P_((const char *, ...));
extern long                   random P_((void));
extern int                    fprintf P_((FILE *, const char *, ...));
extern int                    fflush P_((FILE *));
extern void                   bzero P_((void *, size_t));

#undef P_
#endif

#define TPLANETSPEED    4
#define TPLANETDELAY    15
#define CENTERFIX       40
#define PRANGE		PFIREDIST/4

extern int	nukepl[MAXPLANETS];
extern int	num_nukepl;
extern int      debug;

static struct planet	*k;
static int      	o_clock;
static int      	curr_nuke;
static int		lx,ly;
static void		_move();

void
borgmove()
{
   o_clock++;
   _move();
}

static void
_move()
{
   static int	   dntime;

   if (me->p_status == PFREE)
      exitBorg(-2);
   if (me->p_status == PDEAD)
      exitBorg(-3);

   me->p_ghostbuster = 0;

   if(curr_nuke == num_nukepl){
      if(!dntime) dntime = 15;
      if(--dntime == 0){
	 signal(SIGALRM, SIG_IGN);
	 exitBorg(-4);
      }
   }
   else 
      moveborg();
}

void
moveborg()
{
   unsigned char   crs;
   int             dist;

   k = &planets[nukepl[curr_nuke]];

   if (getplanet(k, &crs, &dist)) {
      if (dist < 5000) {
	 me->p_desspeed = 5;
	 me->p_desdir = crs + 64;
	 if((o_clock % 10) == 0)
	    nukeplanet(k, crs);
	 else if((o_clock % 2) == 0){
	    chk_expltorps();
	    modphaser(k);
	 }
	 else
	    modphaser(k);
      } else if(angdist(crs, me->p_dir) > 16){
	 me->p_desspeed = 6;
	 me->p_desdir = crs;
      }
      else{
	 me->p_desspeed = 9;
	 me->p_desdir = crs;
      }
      if (debug) {
	 printf("%s found planet %s\n", me->p_name,
		k->pl_name);
      }
   }
}

int
opteam(i)

   int	i;
{
   switch(k->pl_owner){
      case FED:	
	 if(i == KPLASMA)
	    return ROM;
	 else
	    return KLI;
	 break;
      case ROM:	
	 if(i == KPLASMA)
	    return FED;
	 else
	    return ORI;
	 break;
      case KLI:	
	 if(i == KPLASMA)
	    return FED;
	 else
	    return ROM;
	 break;
      case ORI:	
	 if(i == KPLASMA)
	    return FED;
	 else
	    return ROM;
	 break;
      default:
	 return 0;
   }
}

void
nukeplanet(k, dir)

   struct planet  *k;
   unsigned char   dir;
{
   int			i;
   if (debug) {
      printf("%s nuking planet %s\n", me->p_name,
	     k->pl_name);
   }
   firephaser(k->pl_x, k->pl_y, dir);
   /*
   posplasma(lx, ly);
   explodepltorp(k->pl_x, k->pl_y);
   */
   fireshrapnel(k->pl_x, k->pl_y);

   i = k->pl_armies - random()%5;
   if(i <= 0){
      k->pl_armies = 0;
      k->pl_owner = 0;
      k->pl_info = (FED | ROM | KLI | ORI);	/* make seen by all */
      curr_nuke ++;
   }
   else
      k->pl_armies = i;
}

void
firephaser(x, y, dir)
   int             x, y;
   unsigned char   dir;
{
   struct phaser  *mine;

   mine = &phasers[me->p_no];
   mine->ph_dir = dir;
   mine->ph_fuse = 10;
   mine->ph_status = PHHIT2;
   mine->ph_x = lx = x - CENTERFIX + (random()%(2*PRANGE)-PRANGE);
   mine->ph_y = ly = y - CENTERFIX + (random()%(2*PRANGE)-PRANGE);
}

void
modphaser(k)
   
   struct planet	*k;
{
   struct phaser  *mine;

   mine = &phasers[me->p_no];
   mine->ph_status = PHHIT2;
   mine->ph_dir++;		/* keep updating */
   mine->ph_x = (k->pl_x - CENTERFIX) + (random()%(2*PRANGE)-PRANGE);
   mine->ph_y = (k->pl_y - CENTERFIX) + (random()%(2*PRANGE)-PRANGE);
   /*
   posplasma(mine->ph_x, mine->ph_y);
   */
   mine->ph_fuse = 10;
   mine->ph_damage = 0;
}

void
posplasma(x, y)
   int             x, y;
{
   struct plasmatorp *t;

   t = &plasmatorps[me->p_no];

   t->pt_no = me->p_no;
   t->pt_status = PTMOVE;
   t->pt_owner = me->p_no;
   t->pt_team = me->p_team;
   t->pt_x = x;
   t->pt_y = y;


   t->pt_damage = 0;
   t->pt_turns = 0;
   t->pt_fuse = 50;
   t->pt_war = ALLTEAM;
   t->pt_speed = 0;

   me->p_nplasmatorp = 1;
}

void
explodepltorp(x, y)
   int             x, y;
{
   struct plasmatorp *t;

   t = &plasmatorps[me->p_no];

   t->pt_x = lx;
   t->pt_y = ly;
   t->pt_status = PTEXPLODE;
   t->pt_fuse = 10;
}

void
fireshrapnel(x, y)
   int             x, y;
{
   register        i;
   register struct torp *k;

   for (i = me->p_no * MAXTORP, k = &torps[i]; 
      i < me->p_no * MAXTORP + MAXTORP; i++, k++) {

      k->t_no = i;
      k->t_owner = me->p_no;
      k->t_status = TMOVE;
      k->t_team = me->p_team;
      k->t_x = lx;
      k->t_y = ly;
      k->t_dir = i * 32;
      k->t_speed = TPLANETSPEED;
      k->t_damage = 0;
      k->t_fuse = TPLANETDELAY;
      k->t_war = 0;
   }
   me->p_ntorp = 8;
}

void
chk_expltorps()
{
   register        i;
   register struct torp *k;

   if((random()%3)==0) return;

   for (i = me->p_no * MAXTORP, k = &torps[i]; 
      i < me->p_no * MAXTORP + MAXTORP; i++, k++) {

      if(k->t_status == TMOVE){
	 k->t_status = TEXPLODE;
	 k->t_fuse = 10;
      }
   }
}

int
getplanet(k, crs, dist)

   struct planet  *k;
   unsigned char  *crs;
   int            *dist;
{
   *dist = ihypot(me->p_x - k->pl_x, me->p_y - k->pl_y);
   *crs = getacourse(k->pl_x, k->pl_y, me->p_x, me->p_y);
   return 1;
}

void
leavemesg()
{
   char            buf[80];
   char            addrbuf[20];

   sprintf(buf, "%s (%s) leaving the dimension (%s)",
	   me->p_name,
	   me->p_mapchars,
	   me->p_login);
   sprintf(addrbuf, " %c%c->ALL",
	   teamlet[me->p_team], shipnos[me->p_no]);
   pmessage2(buf, 0, MALL, addrbuf, me->p_no);
}

void
exitBorg(sig)
   int             sig;
{
   register int    l;
   register struct torp *k;
   register struct plasmatorp *f;

   signal(SIGALRM, SIG_IGN);

   fprintf(stderr, "borg exiting on signal %d\n", sig);
   fflush(stderr);

   for (l = me->p_no * MAXTORP, k = &torps[l];
	l < me->p_no * MAXTORP + MAXTORP; l++, k++) {
      /* torps are free here */
      if (k->t_status != TFREE)
	 k->t_status = TOFF;
   }

   /* plasma */
   f = &plasmatorps[me->p_no * MAXPLASMA];
   f->pt_status = PTFREE;

   bzero(&phasers[me->p_no], sizeof(struct phaser));
   bzero(me, sizeof(struct player));
   me->p_stats.st_tticks = 1;

   /*
    * note: if we had a segmentation violation or bus error we want a core
    * file to debug
    */
   if (sig == SIGBUS || sig == SIGSEGV)
      abort();
   else
      exit(0);
}

void
pmessage(str, recip, group, address)
   char           *str;
   int             recip;
   int             group;
   char           *address;
{
   pmessage2(str, recip, group, address, 255);
}

void
pmessage2(str, recip, group, address, from)
   char           *str;
   int             recip;
   int             group;
   char           *address;
   unsigned char   from;
{
   struct message *cur;
   int             mesgnum;

   if ((mesgnum = ++(mctl->mc_current)) >= MAXMESSAGE) {
      mctl->mc_current = 0;
      mesgnum = 0;
   }
   cur = &messages[mesgnum];
   cur->m_no = mesgnum;
   cur->m_flags = group;
   cur->m_recpt = recip;
   cur->m_from = from;
   (void) sprintf(cur->m_data, "%-9s %s", address, str);
   cur->m_flags |= MVALID;
#ifdef SHORT_PACKETS
   cur->args[0] = DINVALID; /* to prevent false sending with SPS_S_WARNING*/
#endif
}

/* get course from (mx,my) to (x,y) */
unsigned char 
getacourse(x, y, mx, my)
   int             x, y, mx, my;
{
   unsigned char d;
   if (x == mx && y == my)
      return 0;

   d = (unsigned char) (atan2((double) (x - mx),
				 (double) (my - y)) / 3.14159 * 128.);
   MATH_ERR_CHECK(x-mx + my-y);
   return d;
}

/*
 * gives approximate distance from (x1,y1) to (x2,y2) with only
 * overestimations, and then never by more than (9/8) + one bit uncertainty.
 */

int 
ihypot(x1, y1)
   int             x1, y1;
{
   register        x2 = 0, y2 = 0;

   if ((x2 -= x1) < 0)
      x2 = -x2;
   if ((y2 -= y1) < 0)
      y2 = -y2;
   return (x2 + y2 - (((x2 > y2) ? y2 : x2) >> 1));
}
