/*
 * death.c
 */
#include "copyright.h"

#include <stdio.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/types.h>
#include <sys/time.h>
#include "defs.h"
#include "struct.h"
#include "data.h"
#include "packets.h"
#ifndef _NO_DEPEND
#   include "ntserv_protos.h"
#endif

#ifdef USE_SYS_PROTOS
#if defined(__STDC__) || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

extern int                    setitimer P_((int, struct itimerval *, struct itimerval *));
extern int                    printf P_((const char *, ...));
extern int                    fflush P_((FILE *));

#undef P_
#endif


extern jmp_buf  env;

#ifndef DS
static struct itimerval udt;
#endif

void
death()
{
   signal(SIGALRM, SIG_IGN);
#ifndef DS
   udt.it_interval.tv_sec = 0;
   udt.it_interval.tv_usec = 0;
   udt.it_value.tv_sec = 0;
   udt.it_value.tv_usec = 0;
   setitimer(ITIMER_REAL, &udt, 0);
#endif

   me->p_status = POUTFIT;	/* Stop the ghost buster */

#ifdef SB_TRANSWARP
   me->p_flags &= ~PFTWARP;
#endif

   switch (me->p_whydead) {
#ifdef CHAIN_REACTION
   case KTORP2:
   case KSHIP2:
#endif

   case KTORP:
   case KPLASMA:
   case KPHASER:
   case KPLANET:
   case KSHIP:
   case KGENOCIDE:
   case KGHOST:
   case KPROVIDENCE:
      break;
   case KQUIT:
   case KDAEMON:
   case KWINNER:
   default:
#ifndef BASEPRACTICE
      mustexit = 1;
#endif
      break;
   }

   /*
    * if they've had less then 10 minutes of t-mode and galaxy was conquered,
    * don't make them quit.
    */
   if (me->p_whydead == KWINNER &&
       me->p_stats.st_tticks - startTticks < (10 * 60 * (1000000 / UPDATE)))
      mustexit = 0;

   me->p_flags &= ~(PFWAR | PFREFITTING);

#ifdef BASEPRACTICE
   if (me->p_stats.st_rank < NUMRANKS - 1) {
      basepromo();
   } else
#endif

      /* First we check for promotions: */
   if (me->p_stats.st_rank < NUMRANKS - 1) {
      float           ratingTotals;

      ratingTotals = planetRating(me) + bombingRating(me) + offenseRating(me);

      /*
       * The person is promoted if they have enough hours and their ratings
       * are high enough, or if it is clear that they could sit around and do
       * nothing for the amount of time required, and still make the rank.
       * (i.e. their 'DI' rating is about rank.hours * rank.ratings ad they
       * have insufficient time yet).
       */
      if (defenseRating(me) >= ranks[mystats->st_rank + 1].defense &&
	  ((mystats->st_tticks / 36000.0 >= ranks[mystats->st_rank + 1].hours / hourratio &&
	    ratingTotals >= ranks[mystats->st_rank + 1].ratings) ||
	   (mystats->st_tticks / 36000.0 < ranks[mystats->st_rank + 1].hours / hourratio &&
	    ratingTotals * (mystats->st_tticks / 36000.0) >=
	    ranks[mystats->st_rank + 1].hours / hourratio * ranks[mystats->st_rank + 1].ratings))
	 ) {
	 mystats->st_rank++;
	 promotion = 1;
      } else if
	 /*
	  * We also promote if they belong in their current rank, but have
	  * twice enough DI for the next rank.
	  */
	    ((defenseRating(me) >= ranks[mystats->st_rank].defense &&
	      ratingTotals >= ranks[mystats->st_rank].ratings) &&
	     ratingTotals * (mystats->st_tticks / 36000.0) >=
	     ranks[mystats->st_rank + 1].hours / hourratio * ranks[mystats->st_rank + 1].ratings * 2) {
	 mystats->st_rank++;
	 promotion = 1;
      } else if
	 /*
	  * We also promote if they belong down a rank, but have four times
	  * the DI for the next rank.
	  */
	    (mystats->st_rank > 0 &&
	     (defenseRating(me) >= ranks[mystats->st_rank - 1].defense &&
	      ratingTotals >= ranks[mystats->st_rank - 1].ratings) &&
	     ratingTotals * (mystats->st_tticks / 36000.0) >=
	     ranks[mystats->st_rank + 1].hours / hourratio * ranks[mystats->st_rank + 1].ratings * 4) {
	 mystats->st_rank++;
	 promotion = 1;
      }
   }
   /* Give them one more update now... */
   updateClient();
   savestats();
   longjmp(env, 0);
}

#ifdef BASEPRACTICE

static
struct rank     sbranks[NUMRANKS] = {
   {0.0,  0.0, 0.0, "Ensign"},
   {2.0,  0.7, 0.0, "Lieutenant"},
   {4.0,  1.3, 0.0, "Lt. Cmdr."},
   {8.0,  2.0, 0.0, "Commander"},
   {15.0, 2.7, 0.0, "Captain"},	
   {20.0, 3.3, 0.0, "Flt. Capt."},
   {25.0, 4.0, 0.0, "Commodore"},
   {30.0, 4.7, 0.0, "Rear Adm."},
   {40.0, 5.3, 0.0, "Admiral"}};

void
basepromo()
{
   int             myrank = mystats->st_rank;
   double          mytime = (double) mystats->st_tticks / 36000.;
   double          mydi;
   register        i, times;
   double          ratingTotals, dix;

   ratingTotals = damInflictedRating(me) + damRepairedRating(me) +
		  sbThirdRating(me);
   mydi = ratingTotals * mytime;

   /* enough time, enough ratings */
   if ((mytime >= ranks[myrank + 1].hours / hourratio &&
	ratingTotals >= ranks[myrank + 1].ratings) ||

   /* not enough time, enough DI */
       (mytime < ranks[myrank + 1].hours / hourratio &&
       ratingTotals * mytime >= ranks[myrank + 1].hours / hourratio *
       ranks[myrank + 1].ratings)) {

      mystats->st_rank++;
      promotion = 1;

      printf("%s (%c%c) promoted to %s on normal DI.\n",
	     me->p_name, me->p_mapchars[0], me->p_mapchars[1],
	     ranks[mystats->st_rank].name);
      printf("   %f + %f = %f, time %f\n",
	     damInflictedRating(me),
	     damRepairedRating(me),
	     ratingTotals,
	     mytime);
      fflush(stdout);

      return;
   }
   /* di X promotions */

   times = 2;
   i = 0;

   while (myrank + i >= 0) {

      dix = ranks[myrank + 1].hours / hourratio * ranks[myrank + 1].ratings *
	 times;

      if (ratingTotals >= ranks[myrank + i].ratings &&
	  ratingTotals * mytime >= dix) {

	 if (mystats->st_rank < NUMRANKS - 1){	/* failsafe */
	    mystats->st_rank++;
	    promotion = 1;
	 }
	 else
	    return;

	 printf("%s (%c%c) promoted to %s on %d X DI (%f).\n",
		me->p_name, me->p_mapchars[0], me->p_mapchars[1],
		ranks[mystats->st_rank].name,
		times, dix);
	 printf("   %f + %f = %f, time %f\n",
	     damInflictedRating(me),
	     damRepairedRating(me),
	     ratingTotals,
	     mytime);
	 fflush(stdout);

	 return;
      }
      i--;
      times *= 2;
   }
}

#endif				/* BASEPRACTICE */
