/*
 * redraw.c
 */
#include "copyright.h"

#include <stdio.h>
#include <signal.h>
#include <math.h>
#include "defs.h"
#include "struct.h"
#include "data.h"
#include "packets.h"
#ifndef _NO_DEPEND
#   include "ntserv_protos.h"
#endif

#ifdef RSA
#define RSA_CHECK			/* until RSA-only */
#endif

#ifdef USE_SYS_PROTOS
#if defined(__STDC__) || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

extern void                   bcopy P_((const void *, void *, int));
extern int                    fprintf P_((FILE *, const char *, ...));
extern int                    fflush P_((FILE *));
extern time_t                 time P_((time_t *));
extern int                    strncmp P_((const char *, const char *, size_t));
extern int                    sscanf P_((const char *, const char *, ...));
extern int                    fclose P_((FILE *));
extern long                   random P_((void));
extern void                   perror P_((const char *));
extern int 		      shutdown P_((int s, int how));
extern int 		      close P_((int));

#undef P_
#endif

void
intrupt()
{
#ifdef nodef
   /* SHORT_PACKETS TEST */
   printf("interrupt at %d\n", msetime());
#endif
   if (!binconfirm)
      testtime = 0;		/* skip binary confirmation if not enabled */
   else if (bypassed == 1) {
      warning(1, "You have been allowed to bypass the blessed binary check.", 
	      -1);
      testtime = 0;
      binconfirm = 0;
   }
   if (testtime == -1) {
      struct reserved_spacket sp;

      /*
       * Give reasonable period of time to respond to query (and test code if
       * they need to)
       */
      testtime = 200;
      makeReservedPacket(&sp);	/* fill packet w/random numbers */
#ifdef nodef	/* !! sp.data isn't KEY_SIZE large */
      bcopy(sp.data, testdata, KEY_SIZE);	/* copy into testdata */
#else
      bcopy(sp.data, testdata, 16);	/* copy into testdata */
#endif
/*
printf("sending reserved packet to %s@%s\n", me->p_login, me->p_monitor);
fflush(stdout);
*/
      sendClientPacket((struct player_spacket *)&sp);
   } else if (testtime != 0) {
      testtime--;
      if (testtime == 0) {
	 if (binconfirm == 1 || binconfirm == 2) {
	    /* User failed to respond to verification query.  Bye! */
	    me->p_explode = 10 / UPDATE_NORM;
	    me->p_status = PEXPLODE;
	    me->p_whydead = KQUIT;
	    fprintf(stderr, "User binary failed to verify: %s@%s\n",
		    me->p_login, host);	/* changed from me->p_monitor */
	    fflush(stderr);
#ifdef RSA
	    if (RSA_Client == 1)
	       warning(1, "You must use a \"blessed\" binary.", -1);
	    else if (RSA_Client == 2)
	       warning(1, "Wrong RSA client version number.", -1);
	    else
	       warning(1, "You need a spiffy new RSA client for this server.",
		       -1);
#else
	    warning(1, "You must use a \"blessed\" binary.", -1);
#endif
	    /* exit(0); */
	 }
#ifdef CYBORGS
	 else {			/* binconfirm == 3 */
	    int             x = strlen(me->p_name);
	    if (me->p_name[x - 1] != 01) {
	       if (x > 13)
		  x = 13;
	       me->p_name[x] = ' ';
	       me->p_name[x + 1] = 01;
	       me->p_name[x + 2] = '\0';
	    }
	    me->p_flags |= PFCYBORG;
	 }
#endif
      }
   }
   if (me->p_status == PFREE) {
      me->p_ghostbuster = 0;
      me->p_status = PDEAD;
   }
   if (me->p_status == PEXPLODE || me->p_status == PDEAD) {
      inputMask = 0;
   }
   if (((me->p_status == PDEAD) || (me->p_status == POUTFIT))
       && (me->p_ntorp <= 0) && (me->p_nplasmatorp <= 0)) {
      death();
   }
#ifdef COPFIX
   /*
    * In this scheme one of the two players will be booted. I don't know if
    * there is a way to tell, though. The mustexit check is a kludge to so
    * that this loop won't get run for every cycle of the players explosion.
    */
   if (((repCount % 50) == 0) && !mustexit) {
      extern char     start_login[16], start_host[16];	/* from main.c */

      time_t          t = time(NULL);

      if (strncmp(me->p_login, start_login, sizeof(start_login) - 1) != 0 ||
	  strncmp(me->p_monitor, start_host, sizeof(start_host) - 1) != 0) {
#ifdef nodef	/* not sure if this can be done gracefully */
	 me->p_explode = 10 / UPDATE_NORM;
	 me->p_status = PEXPLODE;
	 me->p_whydead = KDAEMON;
	 /*
	  * at this point both players have exploded.  Only this one will be
	  * forced to exit the game though.
	  */
	 warning(1, "Sorry, your slot has been taken by another player.", -1);
	 mustexit = 1;

#endif
	 /*
	  * and print out a message so we'll know if this scheme is working
	  * or not (these are captured by ntstart into startup.err)
	  */

	 fprintf(stderr, "COPILOT: %s@%s became %s@%s -- %s",
		 start_login, start_host, me->p_login, me->p_monitor,
		 ctime(&t));
	 forceShutdown(-12);
      }
   }
#endif

   auto_features();
#if defined(RSA_CHECK) && !defined(BASEPRACTICE)
   rsa_check();
#endif
#ifdef MESG_CHECK
   mesg_check();
#endif

   updateClient();
}

/*
 * These are routines that need to be done on interrupts but don't belong in
 * the redraw routine and particularly don't belong in the daemon.
 */

void
auto_features()
{
   char            buf[180];
   struct player  *pl;
   struct planet  *pln;
   unsigned char   course;
   int             troop_capacity = 0;

   if (me->p_flags & PFSELFDEST) {
      if ((me->p_updates >= selfdest) ||
	  ((me->p_flags & PFGREEN) && (me->p_damage == 0)
	   && (me->p_shield == me->p_ship.s_maxshield))
#ifdef BASEPRACTICE
	  || (me->p_ship.s_type == STARBASE && (me->p_flags & PFGREEN))
#endif
	 ) {
	 me->p_flags &= ~PFSELFDEST;
	 me->p_explode = 10 / UPDATE_NORM;
	 me->p_whydead = KQUIT;
	 me->p_status = PEXPLODE;
      } else {
	 switch ((selfdest - me->p_updates) / 10) {
	 case 5:
	 case 4:
	    warning(1, "You notice everyone on the bridge is staring at you.",
		    83);
	    break;
	 default:
#ifdef SHORT_PACKETS
	    if(send_short){
	       swarning(ONEARG_TEXT, 1, 
		  (unsigned char)((selfdest - me->p_updates) / 10));
	       break;
	    }
#endif
	    sprintf(buf, "Stand By ... Self Destruct in %d seconds",
		    (selfdest - me->p_updates) / 10);
	    warning(1, buf, -1);
	    break;
	 }
      }
   }

    /* provide a refit countdown 4/6/92 TC */

   if (me->p_flags & PFREFITTING) {
      static int lastRefitValue = 0; /* for smooth display */
      if (lastRefitValue != (rdelay - me->p_updates)/10) {
	 lastRefitValue = (rdelay - me->p_updates)/10; /* CSE to the rescue? */

	 switch (lastRefitValue) {
	    case 3:
	    case 2:
	       sprintf(buf, "Engineering:  Energizing transporters in %d seconds", lastRefitValue);
	       warning(1, buf, -1);
	       break;
	    case 1:
	       warning(1, "Engineering:  Energize. [ SFX: chimes ]", -1);
	       break;
	    case 0:
	       switch (random()%5) {
		  case 0:
		     warning(1, "Wait, you forgot your toothbrush!", -1);
		     break;
		  case 1:
		     warning(1, "Nothing like turning in a used ship for a new one.", -1);
		     break;
		  case 2:
		     warning(1, "First officer:  Oh no, not you again... we're doomed!", -1);
		     break;
		  case 3:
		     warning(1, "First officer:  Uh, I'd better run diagnostics on the escape pods.", -1);
		     break;
		  case 4:
		     warning(1, "Shipyard controller:  This time, *please* be more careful, okay?", -1);
		     break;
	       }
               break;
	 }
      }
   }

   /* provide a war declaration countdown 4/6/92 TC */

   if (me->p_flags & PFWAR) {
      static int lastWarValue = 0;
      if (lastWarValue != (delay - me->p_updates)/10) {
	 lastWarValue = (delay - me->p_updates)/10; /* CSE to the rescue? */
	 switch (lastWarValue) {
	    case 9:
	       warning(1, "Weapons officer:  Not again!  This is absurd...", -1);
	       break;
	    case 8:
	       break;
	    case 7:
	       warning(1, "Weapons officer:  ... the whole ship's computer is down?", -1);
	       break;
	    case 6:
	       break;
	    case 5:
	       warning(1, "Weapons officer:  Just to twiddle a few bits of the ship's memory?", -1);
	       break;
	    case 4:
	       break;
	    case 3:
	       warning(1, "Weapons officer:  Bah! [ bangs fist on inoperative console ]", -1);
	       break;
	    case 2:
	       break;
	    case 1:
	       warning(1, "First Officer:  Easy, big guy... it's just one of those mysterious", -1);
	       break;
	    case 0:
	       switch (random()%5) {
		  case 0:
		     warning(1, "First Officer:  laws of the universe, like 'tires on the ether'.", -1);
		     break;
		  case 1:
		     warning(1, "First Officer:  laws of the universe, like 'Klingon bitmaps are ugly'.", -1);
		     break;
		  case 2:
		     warning(1, "First Officer:  laws of the universe, like 'all admirals have scummed'.", -1);
		     break;
		  case 3:
		     warning(1, "First Officer:  laws of the universe, like 'MUCUS PIG exists'.", -1);
		     break;
		  case 4:
		     warning(1, "First Officer:  laws of the universe, like 'guests advance 5x faster'.", -1);
		     break;
	       }
	       break;
	 }
      }
   }

   /* give certain information about bombing or beaming */
   if (me->p_flags & PFBOMB) {
      if (planets[me->p_planet].pl_armies < 5) {
	 int             ac = planets[me->p_planet].pl_armies;
#ifdef SHORT_PACKEST
	 if(send_short){
	    swarning(BOMB_INEFFECTIVE, planets[me->p_planet].pl_armies, 0);
	 }
	 else 
#endif
	 {
	    sprintf(buf, "Weapons Officer: Bombing is ineffective when %d %s %s defending.", ac, ac == 1 ? "army" : "armies", ac == 1 ? "is" : "are");
	    warning(1, buf, -1);
	    me->p_flags &= ~PFBOMB;
	 }
      } else {
#ifdef SHORT_PACKETS
	 if(send_short){
	    swarning(BOMB_TEXT, (unsigned char) me->p_planet,
			     (unsigned char) planets[me->p_planet].pl_armies);
	 }
	 else 
#endif
	 {
	    sprintf(buf, "Weapons Officer: Bombarding %s...  Sensors read %d armies left.",
		 planets[me->p_planet].pl_name,
		 planets[me->p_planet].pl_armies);
	    warning(1, buf, -1);
	 }
      }
   }
   if (me->p_ship.s_type == ASSAULT)
      troop_capacity = (((me->p_kills * 3) > myship->s_maxarmies) ?
			myship->s_maxarmies : (int) (me->p_kills * 3));
   else if (me->p_ship.s_type != STARBASE)
      troop_capacity = (((me->p_kills * 2) > myship->s_maxarmies) ?
			myship->s_maxarmies : (int) (me->p_kills * 2));
   else
      troop_capacity = me->p_ship.s_maxarmies;

   if (me->p_flags & PFBEAMUP) {
      if (me->p_flags & PFORBIT) {
	 if (planets[me->p_planet].pl_armies < 5) {
#ifdef SHORT_PACKETS
	    if(send_short){
	       swarning(BEAMUP_TEXT, (unsigned char)me->p_planet, 0);
	    }
	    else 
#endif
	    {
	       sprintf(buf, "%s: Too few armies to beam up",
		       planets[me->p_planet].pl_name);
	       warning(1, buf, -1);
	    }
	    me->p_flags &= ~PFBEAMUP;
	 } else if (me->p_armies == troop_capacity) {
	    warning(1, "No more room on board for armies", 82);
	    me->p_flags &= ~PFBEAMUP;
	 } else {
#ifdef SHORT_PACKETS
	    if(send_short){
	       swarning(BEAMUP2_TEXT, (unsigned char) me->p_armies,
				   (unsigned char) troop_capacity);
	    }
	    else 
#endif
	    {
	       sprintf(buf, 
		  "Beaming up.  (%d/%d)", me->p_armies, troop_capacity);
	       warning(1, buf, -1);
	    }
	 }
      } else if (me->p_flags & PFDOCK) {
	 if (players[me->p_docked].p_armies == 0) {
#ifdef SHORT_PACKETS
	    if(send_short){
	       swarning(BEAMUPSTARBASE_TEXT, (unsigned char)me->p_docked, 0);
	    }
	    else
#endif
	    {
	       sprintf(buf, "Starbase %s: Too few armies to beam up",
		       players[me->p_docked].p_name);
	       warning(1, buf, -1);
	    }
	    me->p_flags &= ~PFBEAMUP;
	 } else if (me->p_armies == troop_capacity) {
	    warning(1, "No more room on board for armies", 82);
	    me->p_flags &= ~PFBEAMUP;
	 } else {
#ifdef SHORT_PACKETS
	    if(send_short){
	       swarning(BEAMUP2_TEXT, (unsigned char)me->p_armies,
				   (unsigned char) troop_capacity);
	    }
	    else 
#endif
	    {
	       sprintf(buf, 
		  "Beaming up.  (%d/%d)", me->p_armies, troop_capacity);
	       warning(1, buf, -1);
	    }
	 }
      }
   }
   if (me->p_flags & PFBEAMDOWN) {
      if (me->p_armies == 0) {
	 if(me->p_flags & PFDOCK){
#ifdef SHORT_PACKETS
	    if(send_short){
	       swarning(BEAMDOWNSTARBASE_TEXT, (unsigned char)me->p_docked, 0);
	    }
	    else 
#endif
	    {
	       sprintf(buf, "No more armies to beam down to Starbase %s.",
		       players[me->p_docked].p_name);
	       warning(1, buf, -1);
	    }
	 }
	 else {
#ifdef SHORT_PACKETS
	    if(send_short){
	       swarning(BEAMDOWNPLANET_TEXT, (unsigned char)me->p_planet, 0);
	    }
	    else 
#endif
	    {
	       sprintf(buf, "No more armies to beam down to %s.",
		       planets[me->p_planet].pl_name);
	       warning(1, buf, -1);
	    }
	 }
	 me->p_flags &= ~PFBEAMDOWN;
      } else if (me->p_flags & PFORBIT) {
#ifdef SHORT_PACKETS
	 if(send_short){
	    swarning2(ARGUMENTS, (unsigned char) me->p_armies, 
			     (unsigned char) troop_capacity,
	             BEAM_D_PLANET_TEXT, (unsigned char) me->p_planet,
			(unsigned char) planets[me->p_planet].pl_armies);
	 }
	 else 
#endif
	 {
	    sprintf(buf, "Beaming down.  (%d/%d) %s has %d armies left",
		    me->p_armies,
		    troop_capacity,
		    planets[me->p_planet].pl_name,
		    planets[me->p_planet].pl_armies);
	    warning(1, buf, -1);
	 }
      } else if (me->p_flags & PFDOCK) {
	 if (players[me->p_docked].p_armies ==
	     players[me->p_docked].p_ship.s_maxarmies) {
#ifdef SHORT_PACKETS
	    if(send_short){
	       swarning(SBREPORT, (unsigned char) me->p_docked, 0);
	    }
	    else 
#endif
	    {
	       sprintf(buf, "Transporter Room:  Starbase %s reports all troop bunkers are full!",
		       players[me->p_docked].p_name);
	       warning(1, buf, -1);
	    }
	    me->p_flags &= ~PFBEAMDOWN;
	 } else {
#ifdef SHORT_PACKETS
	    if(send_short){
	       swarning2(ARGUMENTS, (unsigned char) me->p_armies, 
				(unsigned char) troop_capacity,
	                 BEAM_U_TEXT, (unsigned char) me->p_docked,
			   (unsigned char) players[me->p_docked].p_armies);
	    }
	    else 
#endif
	    {
	       sprintf(buf, "Transfering ground units.  (%d/%d) Starbase %s has %d armies left",
		    me->p_armies,
		    troop_capacity,
		    players[me->p_docked].p_name,
		    players[me->p_docked].p_armies);
	       warning(1, buf, -1);
	    }
	 }
      }
   }
   if (me->p_flags & PFREPAIR) {
      if ((me->p_damage == 0) && (me->p_shield == me->p_ship.s_maxshield))
	 me->p_flags &= ~PFREPAIR;
   }
   if (me->p_flags & PFPLOCK) {	/* set course to player x */
      int             dist;

      pl = &players[me->p_playerl];

      /* isae - break the lock if he's cloaked */
      if (pl->p_status != PALIVE || (pl->p_flags & PFCLOAK))
	 me->p_flags &= ~PFPLOCK;

      if (pl->p_ship.s_type == STARBASE) {
#ifdef SB_TRANSWARP
	 if (me->p_flags & PFTWARP){
	    if (pl->p_status != PALIVE){
	       warning(1, "Starbase is dead; Transwarp aborted!", -1);
	       me->p_flags &= ~PFTWARP;
	       me->p_speed = 3;
	       set_speed(0);
	    }
	    if(me->p_etemp > me->p_ship.s_maxegntemp - 10){
	       warning(1, "Warp drive reaching critical temperature!", -1);
	       me->p_flags &= ~PFTWARP;
	       me->p_speed = 3;
	       set_speed(0);
	    }
	    else
	    if (me->p_fuel < (int) (me->p_ship.s_maxfuel/4)){
	       warning(1, "Not enough fuel to continue transwarp!", -1);
	       me->p_flags &= ~PFTWARP;
	       me->p_speed = 3;
	       set_speed(0);
	    }
	    else if(!(me->p_flags & PFPLOCK)){
	       warning(1, "Lost player lock!", -1);
	       me->p_flags &= ~PFTWARP;
	       me->p_speed = 3;
	       set_speed(0);
	    }
	 }
#endif
	 dist = hypot((double) (me->p_x - pl->p_x),
		      (double) (me->p_y - pl->p_y));
#ifdef SB_TRANSWARP
	 if (!(me->p_flags & PFTWARP))
#endif
	 if (dist - (DOCKDIST / 2) < (11500 * me->p_speed * me->p_speed) /
	     me->p_ship.s_decint) {
	    if (me->p_desspeed > 2) {
	       set_speed(me->p_desspeed - 1);
	    }
	 }
#ifdef SB_TRANSWARP
	 if ((dist < 2*DOCKDIST) && (me->p_flags & PFTWARP)){
	    me->p_x = pl->p_x; me->p_y = pl->p_y;
	    me->p_speed=2;
	    me->p_flags &= ~(PFPLOCK);
	    orbit();
	    me->p_flags &= ~(PFTWARP);
	 } else
#endif
	    if ((dist < DOCKDIST) && (me->p_speed <= 2)) {
	       me->p_flags &= ~PFPLOCK;
	       orbit();
	    }
      }
      if (me->p_flags & PFPLOCK) {
	 course = newcourse(pl->p_x, pl->p_y);
#ifdef SB_TRANSWARP
          if (me->p_flags & PFTWARP)
              me->p_dir = course;
          else
#endif
	    set_course(course);
      }
   }
   if (me->p_flags & PFPLLOCK) {/* set course to planet x */
      int             dist;

      pln = &planets[me->p_planet];
      dist = hypot((double) (me->p_x - pln->pl_x),
		   (double) (me->p_y - pln->pl_y));

      /*
       * This is magic.  It should be based on defines, but it slows the ship
       * down to warp two an appropriate distance from the planet for orbit
       */

      if (dist - (ORBDIST / 2) < (11500 * me->p_speed * me->p_speed) /
	  me->p_ship.s_decint * UPDATE_NORM) {
	 if (me->p_desspeed > 2) {
	    set_speed(2);
	 }
      }
      if ((dist < ENTORBDIST) && (me->p_speed <= 2)) {
	 me->p_flags &= ~PFPLLOCK;
	 orbit();
      } else {
	 course = newcourse(pln->pl_x, pln->pl_y);
	 set_course(course);
      }
   }
}

unsigned char
newcourse(x, y)
   int             x, y;
{
   unsigned char	d;
   if(x == me->p_x && y == me->p_y) return 0;

   d = ((unsigned char) (atan2((double) (x - me->p_x),
				  (double) (me->p_y - y)) / 3.14159 * 128.));
   MATH_ERR_CHECK(x-me->p_x + me->p_y-y);
   return d;
}

#ifdef RSA_CHECK
void
rsa_check()
{
   static int	warned;
   static int	rep=1;
   char		buf[80];

   if(testtime > 0 || testtime < 0)
      return;

   if(RSA_Client)
      return;

   if(warned)
      return;

   if((rep++ % 50)!=0)
      return;

   warning(1, "You have very important messages, Captain!", -1);
   bounce(
      "************* READ THIS ** READ THIS ** READ THIS ******************", 
      me->p_no);
   bounce(
      "You don't seem to be using an RSA client. Since this server will be", 
      me->p_no);
   bounce(
      "going RSA-only soon it's important that you get one in order to",
      me->p_no);
   bounce(
      "continue playing here.  Please see the motd (startup screen) for",
      me->p_no);
   bounce(
      "details.  This has been an automatic message.", 
      me->p_no);
   bounce(
      "********************************************************************",
      me->p_no);

   sprintf(buf, "%c%c is not using an RSA client!",
      me->p_mapchars[0], me->p_mapchars[1]);
   pmessage2(buf, 0, MALL, "GOD->ALL", me->p_no);
   
   warned = 1;
}
#endif
