/*
 * torp.c
 */
#include "copyright.h"

#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#include <sys/timeb.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include "defs.h"
#include "struct.h"
#include "data.h"
#include "packets.h"
#ifndef _NO_DEPEND
#   include "ntserv_protos.h"
#endif

#ifdef USE_SYS_PROTOS
#if defined(__STDC__) || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

extern int ftime P_((struct timeb *tp));
extern void                   perror P_((const char *));

#undef P_
#endif

/* Launch torp */

/*
 * * If a set of given conditions are met, fire a single torp in direction *
 * course.  Type is used because robots are allowed to shoot straight. *
 * Torps sent with status TMOVE wobble a bit.  TSTRAIGHT torps (fired * by
 * robots) move in a straight line. *
 * 
 * torp->t_fuse is the life of the torpedo.  It is set here based on * a
 * random function.  Torps currently live two to five seconds.
 */

static struct timeb lasttime;
static struct timeb curtime;
#define GAP 100
long            random();

void
ntorp(course, type)
   unsigned char   course;
   int             type;
{
   register int    i;
   register struct torp *k;
   int             dCourse;

   if (me->p_flags & PFWEP) {
      warning(0, "Torpedo launch tubes have exceeded maximum safe temperature!", 25);
      return;
   }
   ftime(&curtime);
   if ((curtime.time == lasttime.time && curtime.millitm < lasttime.millitm + GAP) ||
       (curtime.time == lasttime.time + 1 && curtime.millitm + 1000 - lasttime.millitm < GAP)) {
      return;
   } else {
      lasttime.time = curtime.time;
      lasttime.millitm = curtime.millitm;
   }

   if (me->p_ntorp == MAXTORP) {
      warning(0, "Our computers limit us to having 8 live torpedos at a time captain!", 26);
      return;
   }
   if (me->p_fuel < myship->s_torpcost) {
      warning(0, "We don't have enough fuel to fire photon torpedos!", 27);
      return;
   }
   if (me->p_flags & PFREPAIR) {
      warning(1, "We cannot fire while our vessel is in repair mode.", 28);
      return;
   }
   if ((me->p_cloakphase > (CLOAK_PHASES - 1) - cloakdelay) &&
       (me->p_ship.s_type != ATT)) {
      warning(0, "We are unable to fire while in cloak, captain!", 29);
      return;
   }
   me->p_ntorp++;
   me->p_fuel -= myship->s_torpcost;
   dCourse = abs(course - me->p_dir);
   if (backTorpHeat && me->p_ship.s_type != STARBASE && ((dCourse > 115) && (dCourse < 145))) {	/* if the shot is backwards (within 30 degrees) and the
                 * global backTorpHeat is on, then heat the weapons */

      me->p_wtemp += (myship->s_torpcost / 10) + (10 * me->p_speed);	/* Heat weapons */
   } else {
      me->p_wtemp += (myship->s_torpcost / 10) - 10;	/* Heat weapons */
   }

   for (i = me->p_no * MAXTORP, k = &torps[i];	/* Find a free torp */
	i < me->p_no * MAXTORP + MAXTORP; i++, k++) {
      if (k->t_status == TFREE)
	 break;
   }

   /* Setup data in new torp */
   if (type > TSTRAIGHT || type < TFREE)
      type = TMOVE;
   if (straightorps)
      type = TSTRAIGHT;		/* jfy */
   k->t_no = i;
   k->t_status = type;
   k->t_owner = me->p_no;
   k->t_team = me->p_team;
   k->t_x = me->p_x;
   k->t_y = me->p_y;
   k->t_dir = course;
   k->t_damage = myship->s_torpdamage;
   if (vectortorps)
      k->t_speed = vector_torp_speed(me->p_dir, me->p_speed, course,
				     myship->s_torpspeed);
   else
      k->t_speed = myship->s_torpspeed;
   k->t_war = me->p_hostile | me->p_swar;
   k->t_fuse = myship->s_torpfuse + (random() % 20);
   k->t_turns = myship->s_torpturns;
#ifdef MINES
   k->t_mine = 0;
#endif
#ifdef CHAIN_REACTION
   k->t_whodet = NODET;  /* isae - mark it as undetted */
#endif
}

int
vector_torp_speed(pdir, pspeed, tdir, tspeed)
   unsigned char   pdir, tdir;
   int             pspeed, tspeed;
{
   double          vsx, vsy, vtx, vty, ntx, nty, sqrt();
   int             newspeed;

   /* ship vector */
   vsx = (double) pspeed *Cos[pdir] * WARP1;
   vsy = (double) pspeed *Sin[pdir] * WARP1;
   /* torp vector */
   vtx = (double) tspeed *Cos[tdir] * WARP1;
   vty = (double) tspeed *Sin[tdir] * WARP1;

   ntx = vsx + vtx;
   nty = vsy + vty;

   newspeed = nint(sqrt(ntx * ntx + nty * nty));
   MATH_ERR_CHECK((int)(ntx*ntx+nty*nty));
   newspeed = newspeed / WARP1;
   /*
    * printf("vectortorp: ship (s=%d,d=%d,t=%d), result: %d\n", pspeed, pdir,
    * tspeed, newspeed); fflush(stdout);
    */
   if (newspeed > 20)
      newspeed = 20;


   return newspeed;
}
