#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <errno.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include "defs.h"
#include "struct.h"
#include "data.h"
#include "packets.h"

#if __STDC__ || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

#ifdef USE_SYS_PROTOS
extern void                   perror P_((const char *));
#endif

#undef P_

int		num_db_entries;
char		**db_entries;

void
main(argc, argv)
   
   int	argc;
   char	**argv;
{
   int			plfd, plfd_db;
   int			pos;
   struct statentry	player;

   if(!argc) {
      fprintf(stderr, "usage: %s <merge name> ... <into name>\n",
	 argv[0]);
      exit(0);
   }


   

   plfd = open(PLAYERFILE, O_RDONLY, 0664);
   if(plfd < 0){
      perror(PLAYERFILE);
      exit(1);
   }
   plfd_db = open(PLAYERFILE_DB, O_RDWR | O_CREAT, 0644);
   if(plfd_db < 0){
      perror(PLAYERFILE_DB);
      exit(1);
   }
   readin_positions(plfd_db);

   while(read(plfd, (char *) &player, sizeof(struct statentry)) ==
	 sizeof(struct statentry)){
      pos = get_position_from_db(&player);
      /* all stat values in PLAYERFILE must be written to 
	 PLAYERFILE_DB.  If 'pos' valid, then this player is written
	 to position 'pos' (overwriting anything old value there).
	 If negative, this player is appended to PLAYERFILE_DB */
      saveplayer(plfd_db, pos, &player);
   }
   close(plfd_db);
   close(plfd);

   printf("All players in `%s' have been saved in `%s'.\n",
      PLAYERFILE, PLAYERFILE_DB);
   printf("`%s' may be now be deleted if the daemon isn't running.\n",
      PLAYERFILE);
}

void
saveplayer(plfd_db, pos, player)

   int			plfd_db;
   int			pos;
   struct statentry	*player;
{
   if(pos > -1){
      /* seek from start of file */
      if(lseek(plfd_db, pos * sizeof(struct statentry), SEEK_SET) < 0){
	 perror("lseek");
	 exit(1);
      }
   }
   else{
      /* append */
      if(lseek(plfd_db, 0, SEEK_END) < 0){
         perror("lseek");
         exit(1);
      }
   }

   /* append if pos was negative */
   if(write(plfd_db, (char *) player, sizeof(struct statentry)) < 0){
      perror("write");
      exit(1);
   }
}
      
int
get_position_from_db(player)

   struct statentry     *player;
{
   register	i;
   for(i=0; i< num_db_entries; i++){
      if(strcmp(player->name, db_entries[i])==0)
	 return i;
   }
   return -1;
}

void
readin_positions(plfd_db)

   int	plfd_db;
{
   struct stat		buf;
   register int		i;
   struct statentry	pl;

   fstat(plfd_db, &buf);

   num_db_entries = buf.st_size / sizeof(struct statentry);
   db_entries = (char **) malloc(sizeof(char *) * num_db_entries);
   for(i=0; i< num_db_entries; i++){
      if(read(plfd_db, (char *) &pl, sizeof(struct statentry)) !=
	 sizeof(struct statentry)){
	 fprintf(stderr, "%s: Couldn't read %d entries. Failed on %d.\n",
	    PLAYERFILE_DB, num_db_entries, i);
	 exit(1);
      }
      db_entries[i] = strdup(pl.name);
   }
}
