#include <stdio.h>
#include <sys/types.h>
#include <sys/file.h>
#include <time.h>
#include <pwd.h>
#include "defs.h"
#include "struct.h"

#define SBRATE(x)	((double) x->sbkills / \
			(x->sblosses + 2.0 + (float)x->sbtime / 14400))

#ifdef BASEPRACTICE
#define NORMALIZED_SBRATE(x)	(SBRATE(x)/ ((double) status->kills /\
	   (status->losses + 2.0 + (double)(status->timeprod/10.)/14400)))
#endif

struct rawdesc {
   struct rawdesc *next;
   char            name[16];
   int             ttime;
   int             rank;
   float           bombing, planets, offense, defense;
   int             nbomb, nplan, nkill, ndie;
   int             sbkills, sblosses, sbtime;
   float           sbmaxkills;
   long            lastlogin;
#ifdef BASEPRACTICE
   double	   dami, damr;
   double	   dam_inf, dam_rep;
#endif
};

#if defined(__STDC__) || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

/* scores.c */
void addNewPlayer P_((char *name, struct player *pl, int mode));
int ordered P_((struct rawdesc *num1, struct rawdesc *num2, int mode));
void fixkeymap P_((char *s));
void usage P_((void));

#ifdef USE_SYS_PROTOS
extern int                    atoi P_((const char *));
extern int                    printf P_((const char *, ...));
extern int                    read P_((int, void *, unsigned int));
extern int                    close P_((int));
extern void                   perror P_((const char *));
extern time_t                 time P_((time_t *));
extern char *                 ctime P_((const time_t *));
extern char *                 strcat P_((char *, const char *));
extern void *                 malloc P_((size_t));
#endif

#undef P_

char           *ranks[] = {
   "Ensigns",
   "Lieutenants",
   "Lieutenant Commanders",
   "Commanders",
   "Captains",
   "Fleet Captains",
   "Commodores",
   "Rear Admirals",
"Admirals"};

/* There is a time bias towards people with at least 4 hours of play... */
#define TIMEBIAS 4

int             daysold = 28;

struct statentry player;
struct status  *status;
struct rawdesc *output = NULL;
char            mode;
int             showall;

void
addNewPlayer(name, pl, mode)
   char           *name, mode;
   struct player  *pl;
{
   struct rawdesc *temp;
   struct rawdesc **temp2;

   temp = (struct rawdesc *) malloc(sizeof(struct rawdesc));
   strcpy(temp->name, name);
   temp->lastlogin = pl->p_stats.st_lastlogin;
   temp->ttime = pl->p_stats.st_tticks / 10;
   temp->rank = pl->p_stats.st_rank;
   temp->bombing = bombingRating(pl);
   temp->planets = planetRating(pl);
   temp->offense = offenseRating(pl);
   temp->defense = defenseRating(pl);
   temp->nbomb = pl->p_stats.st_tarmsbomb;
   temp->nplan = pl->p_stats.st_tplanets;
   temp->nkill = pl->p_stats.st_tkills;
   temp->ndie = pl->p_stats.st_tlosses;
   temp->sbtime = pl->p_stats.st_sbticks / 10;
   temp->sbkills = pl->p_stats.st_sbkills;
   temp->sblosses = pl->p_stats.st_sblosses;
   temp->sbmaxkills = pl->p_stats.st_sbmaxkills;
#ifdef BASEPRACTICE
   temp->dami = damInflictedRating(pl);
   temp->damr = damRepairedRating(pl);
   temp->dam_inf = pl->p_stats.st_dami;
   temp->dam_rep = pl->p_stats.st_damr;
#endif
   temp2 = &output;
   for (;;) {
      if ((*temp2) == NULL || ordered(temp, *temp2, mode)) {
	 temp->next = (*temp2);
	 (*temp2) = temp;
	 break;
      }
      temp2 = &((*temp2)->next);
   }
}

int
ordered(num1, num2, mode)
   struct rawdesc *num1, *num2;
   char            mode;
{
   double           temp, temp2;

   switch (mode) {
   case 'o':
      temp = num1->ttime * (num1->offense) / (num1->ttime / 90 + 40 * TIMEBIAS);
      temp2 = num2->ttime * (num2->offense) / (num2->ttime / 90 + 40 * TIMEBIAS);

      if (temp >= temp2) {
	 return (1);
      } else {
	 return (0);
      }
      break;

   case 'd':
      temp = (float) num1->ttime / (num1->ndie + num1->ttime / 90 + 40 * TIMEBIAS);
      temp2 = (float) num2->ttime / (num2->ndie + num2->ttime / 90 + 40 * TIMEBIAS);

      if (temp >= temp2) {
	 return (1);
      } else {
	 return (0);
      }
      break;

   case 'b':
      temp = num1->ttime * (num1->bombing) / (num1->ttime / 90 + 40 * TIMEBIAS);
      temp2 = num2->ttime * (num2->bombing) / (num2->ttime / 90 + 40 * TIMEBIAS);

      if (temp >= temp2) {
	 return (1);
      } else {
	 return (0);
      }
      break;

   case 'p':
      temp = num1->ttime * (num1->planets) / (num1->ttime / 90 + 40 * TIMEBIAS);
      temp2 = num2->ttime * (num2->planets) / (num2->ttime / 90 + 40 * TIMEBIAS);

      if (temp >= temp2) {
	 return (1);
      } else {
	 return (0);
      }
      break;

   case 'h':
      if (num1->ttime > num2->ttime)
	 return 1;
      else
	 return 0;
      break;

   case 'a':
      if (num1->rank > num2->rank)
         return (1);
      if (num2->rank > num1->rank)
         return (0);
      /* NOTE: fallthrough to DI for equal ranks */

      /*
       * 4/2/89  Rank by DI within class.
       */
   case 'D':
      temp = num1->ttime * (num1->planets + num1->bombing + num1->offense);
      temp2 = num2->ttime * (num2->planets + num2->bombing + num2->offense);

      if (temp >= temp2) {
	 return (1);
      } else {
	 return (0);
      }
      break;

#ifdef BASEPRACTICE
   case 'S':
      temp = num1->dami + num1->damr +
	     (((num1->dami*num1->damr)/2)/TRAD_IMPORTANCE) * NORMALIZED_SBRATE(num1);
/*
printf("%s: %f\n", num1->name, temp);
*/

      temp2 = num2->dami + num2->damr +
	     (((num2->dami*num2->damr)/2)/TRAD_IMPORTANCE) * NORMALIZED_SBRATE(num2);
/*
printf("%s: %f\n", num2->name, temp2);
*/
      return temp > temp2;

#ifdef nodef
      if(num1->dami + num1->damr + ((num1->dami+num1->damr)/3. *num1->defense)> 
	 num2->dami + num2->damr + ((num2->dami+num2->damr)/3. *num2->defense))
	 return 1;
      else
	 return 0;
#endif
      break;
   case 'T':
      if(num1->ttime > num2->ttime)
	 return 1;
      else 
	 return 0;
      break;
#endif

   case 's':
      /*
       * A starbase is ranked by how many kills he gets per death. (roughly).
       * More accurately, his ratio is his rating, where his ratio is
       * kills/(deaths + 2 + hours/4).  So, he is auto credited with 1 death
       * per four hours, and he gets 2 automatically.  This encourages some
       * amount of cowardise in a starbase, but not _too_ much.  It seems to
       * make sense to me ('course, I may be screwed, but that's the way it
       * goes).
       */

      temp = SBRATE(num1);
      temp2 = SBRATE(num2);

      if (temp >= temp2) {
	 return (1);
      } else {
	 return (0);
      }
      break;

   default:
      return (1);
   }
}

void
main(argc, argv)
   int             argc;
   char          **argv;
{
   int             fd;
   struct player   j;
   struct rawdesc *temp;
   int             i;
   long            totalTime = 0;
   char            tmpbuf[100];
   char           *plfile;

   status = (struct status *) malloc(sizeof(struct status));

   showall = 0;
   mode = 'a';

#ifdef LARGE_DB
   plfile = PLAYERFILE_DB;
#else
   plfile = PLAYERFILE;
#endif
   for (i = 1; i < argc; i++) {
      if (argv[i][0] == '-') {
	 switch (argv[i][1]) {
	 case 'f':
	    if (argv[i + 1] == NULL) {
	       usage();
	    }
	    plfile = argv[i + 1];
	    i++;
	    break;
	 default:
	    usage();
	    break;
	 }
      } else {
	 mode = argv[i][0];

	 if (argv[i][1] == 'a') {
	    showall = 1;
	 }
	 if (mode == 'O') {
	    if (argv[i + 1]) {
	       daysold = atoi(argv[i + 1]);
	       i++;
	    }
	 }
      }
   }

   fd = open(GLOBAL, O_RDONLY, 0777);
   if (fd < 0) {
      printf("Cannot open the global file!\n");
      exit(0);
   }
   read(fd, (char *) status, sizeof(struct status));
   close(fd);
   fd = open(plfile, O_RDONLY, 0777);
   if (fd < 0) {
#ifdef LARGE_DB
      perror(PLAYERFILE_DB);
#else
      perror(PLAYERFILE);
#endif
      exit(1);
   }
   switch (mode) {
   case 'O':
   case 'o':
   case 'r':
   case 'b':
   case 'p':
   case 'd':
   case 'a':
   case 'h':
      printf("    Name               Hours      Planets     Bombing     Offense     Defense\n");
      printf("    Totals           %7.2f %11d %11d %11d %11d %f\n",
	     status->time / 36000.0, status->planets, status->armsbomb,
	     status->kills, status->losses, status->timeprod / 10);
#ifdef BASEPRACTICE
      printf("Dam Infl  Dam Rpr'd\n");
      printf("%12.2f %12.2f\n",
	       status->dami,
	       status->damr);
#endif

      break;
   case 'D':
      printf("    Name               Hours   Ratings   DI    Last login\n");
      break;
#ifdef BASEPRACTICE
   case 'S':
      printf("    Name               Hours  Kills  Losses  Ratio Maxkills Dam Infl Dam Rpr'd\n");
      break;
   case 'T':
      printf("    Name               Hours     Dam Infl    Dam Rpr'd   Defense\n");
      break;
#endif
   case 's':
      printf("    Name               Hours   Kills   Losses   Ratio  Maxkills\n");
      break;
   case 't':
      totalTime = status->timeprod / 10;
      break;
   default:
      usage();
      break;
   case 'g':
      printf("                                Non-Tournament                               Tournament                        Starbase\n");
      printf("Name             seconds kills losses planets armies maxkills   seconds kills losses planets armies   seconds kills losses maxkills\n");
      break;
   case 'A':
      printf("%10ld %10d %10d %10d %10d %10f\n",
	     status->time,
	     status->planets,
	     status->armsbomb,
	     status->kills,
	     status->losses,
	     status->timeprod);
      break;
   }

   i = -1;
   while (read(fd, (char *) &player, sizeof(struct statentry)) == sizeof(struct statentry)) {
      i++;
      j.p_stats = player.stats;
      if (j.p_stats.st_tticks < 1)
	 j.p_stats.st_tticks = 1;
      if (player.stats.st_tticks < 1)
	 player.stats.st_tticks = 1;
      fixkeymap(player.stats.st_keymap);
      switch (mode) {
      case 'o':
      case 'b':
      case 'p':
      case 'd':
      case 'a':
      case 'h':
      case 'D':
#ifdef nodef
	 /* Throw away entries with less than 1 hour play time */
	 if (!showall && player.stats.st_tticks < 36000)
	    continue;
#endif
	 /* Throw away entries with less than 8 hour play time */
	 if (!showall && player.stats.st_tticks < 8 * 36000)
	    continue;
      case 's':
#ifdef BASEPRACTICE
      case 'S':
      case 'T':
#endif
#ifdef nodef
	 /* Throw away entries unused for over 4 weeks */
	 if (!showall && time(NULL) > player.stats.st_lastlogin + 2419200)
	    continue;
#endif
#ifdef BASEPRACTICE
	 /* Throw away SB's with less than 1 hour play time */
	 if (!showall && mode == 'S' && player.stats.st_sbticks < 36000)
	    continue;
#else
	 /* Throw away SB's with less than 5 hours play time */
	 if (!showall && mode == 's' && player.stats.st_sbticks < 5*36000)
	    continue;
#endif
	 addNewPlayer(player.name, &j, mode);
	 break;
      case 'O':
	 if (player.stats.st_lastlogin + daysold * 86400 > time(NULL))
	    break;
      case 'r':
	 printf("%3d %-16.16s %7.2f %6d %5.2f %5d %5.2f %5d %5.2f %5d %5.2f\n",
		i,
		player.name,
		player.stats.st_tticks / 36000.0,
		player.stats.st_tplanets,
		planetRating(&j),
		player.stats.st_tarmsbomb,
		bombingRating(&j),
		player.stats.st_tkills,
		offenseRating(&j),
		player.stats.st_tlosses,
		defenseRating(&j));
	 break;
      case 't':
	 totalTime += player.stats.st_ticks / 10 +
	    player.stats.st_sbticks / 10;
	 break;
      default:
      case 'g':
	 printf("%-16.16s %7d %5d %6d %7d %6d %8.2f   %7d %5d %6d %7d %6d   %7d %5d %6d %8.2f\n",
		player.name,
		player.stats.st_ticks / 10,
		player.stats.st_kills,
		player.stats.st_losses,
		player.stats.st_planets,
		player.stats.st_armsbomb,
		player.stats.st_maxkills,
		player.stats.st_tticks / 10,
		player.stats.st_tkills,
		player.stats.st_tlosses,
		player.stats.st_tplanets,
		player.stats.st_tarmsbomb,
		player.stats.st_sbticks / 10,
		player.stats.st_sbkills,
		player.stats.st_sblosses,
		player.stats.st_sbmaxkills);
	 break;
      case 'A':
	 strcpy(tmpbuf, player.name);
	 strcat(tmpbuf, "_");
	 printf("%-16.16s %-16.16s %-96.96s %1d %9.2f %7d %7d %7d %7d %7d %7d %7d %7d %7d %7d %7d %7d %7d %9.2f %9ld %7d\n",
		tmpbuf,
		player.password,
		player.stats.st_keymap,
		player.stats.st_rank,
		player.stats.st_maxkills,
		player.stats.st_kills,
		player.stats.st_losses,
		player.stats.st_armsbomb,
		player.stats.st_planets,
		player.stats.st_ticks,
		player.stats.st_tkills,
		player.stats.st_tlosses,
		player.stats.st_tarmsbomb,
		player.stats.st_tplanets,
		player.stats.st_tticks,
		player.stats.st_sbkills,
		player.stats.st_sblosses,
		player.stats.st_sbticks,
		player.stats.st_sbmaxkills,
		player.stats.st_lastlogin,
		player.stats.st_flags);
	 break;
      }
   }
   close(fd);
   if (mode == 't') {
      printf("Total play time for this game has been at least %ld seconds\n",
	     totalTime);
   }
   if (mode == 'o' || mode == 'b' || mode == 'p' || mode == 'd' || mode == 'a' || mode == 'h' || mode == 's' || mode == 'D' || mode == 'S' || mode == 'T') {
      int             lastrank = -1;
      temp = output;
      i = 0;
      while (temp != NULL) {
	 if ((mode == 'a') && lastrank != temp->rank) {
	    lastrank = temp->rank;
	    printf("\n%s:\n", ranks[lastrank]);
	 }
	 i++;
	 switch (mode) {
	 case 'o':
	 case 'b':
	 case 'p':
	 case 'd':
	 case 'a':
	 case 'h':
	 default:
	    printf("%3d %-16.16s %7.2f %6d %5.2f %5d %5.2f %5d %5.2f %5d %5.2f\n",
		   i,
		   temp->name,
		   temp->ttime / 3600.0,
		   temp->nplan,
		   temp->planets,
		   temp->nbomb,
		   temp->bombing,
		   temp->nkill,
		   temp->offense,
		   temp->ndie,
		   temp->defense);
	    break;
	 case 'D':
	    printf("%3d %-16.16s %7.2f %7.2f %8.2f  %s",
		   i,
		   temp->name,
		   temp->ttime / 3600.0,
		   temp->bombing + temp->offense + temp->planets,
		   (temp->bombing + temp->offense + temp->planets) * temp->ttime / 3600.0,
		   ctime(&(temp->lastlogin)));
	    break;
#ifdef BASEPRACTICE
	 case 'S':
	    printf("%3d %-16.16s %7.2f%7d%8d%7.2f%9.2f  %5.2f    %5.2f\n",
		   i,
		   temp->name,
		   temp->ttime / 3600.0,	/* not sbtime */
		   temp->sbkills,
		   temp->sblosses,
		   (temp->sblosses == 0) ? (float) temp->sbkills : (float) temp->sbkills / temp->sblosses,
		   temp->sbmaxkills,

		   temp->dami,
		   temp->damr);
	    break;
	 
	 case 'T':
	    printf("%3d %-16.16s %7.2f %12.2f %12.2f %12.2f\n",
	       i,
	       temp->name,
	       temp->ttime / 3600.0,
	       temp->dam_inf,
	       temp->dam_rep,
	       temp->defense);
	    break;
#endif
	 case 's':
	    printf("%3d %-16.16s %7.2f %7d %8d %7.2f %9.2f\n",
		   i,
		   temp->name,
		   temp->sbtime / 3600.0,
		   temp->sbkills,
		   temp->sblosses,
		   (temp->sblosses == 0) ? (float) temp->sbkills : (float) temp->sbkills / temp->sblosses,
		   temp->sbmaxkills);
	    break;

	 }
	 temp = temp->next;
      }
   }
   exit(0);
}

void
fixkeymap(s)
   char           *s;
{
   int             i;

   for (i = 0; i < 95; i++) {
      if (s[i] == 0) {
	 s[i] = i + 32;
      }
   }
   s[95] = 0;
}

void
usage()
{
   printf("usage: scores [mode] [-f player file]\n");
   printf("Modes:\n");
   printf("a - print best players in order\n");
   printf("o - list players with best offense\n");
   printf("d - list players with best defense\n");
   printf("h - list players with most hours\n");
   printf("p - list players with best planet rating\n");
   printf("b - list players with best bombing rating\n");
   printf("s - list best starbase players\n");
   printf("S - list best starbase players (base practice server)\n");
   printf("  Note: for modes [aodpbs], players with less than 1 hour\n");
   printf("    or who haven't played for four weeks are omitted\n");
   printf("r - list all players, hours, and ratings (no order)\n");
   printf("A - list entire database in ascii form (for use with newscores)\n");
   printf("t - print rough number of seconds of play time\n");
   printf("O - print players who haven't logged in for 4 weeks\n");
   printf("none: list all players, and their totals\n");
   exit(0);
}
